import 'dart:async';
import 'dart:math';

import 'package:equatable/equatable.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:hydrated_bloc/hydrated_bloc.dart';

import 'package:colors/models/game/game.dart';
import 'package:colors/models/settings/settings_game.dart';
import 'package:colors/models/settings/settings_global.dart';

part 'game_state.dart';

class GameCubit extends HydratedCubit<GameState> {
  GameCubit()
      : super(GameState(
          currentGame: Game.createEmpty(),
        ));

  void updateState(Game game) {
    emit(GameState(
      currentGame: game,
    ));
  }

  void refresh() {
    final Game game = Game(
      // Settings
      gameSettings: state.currentGame.gameSettings,
      globalSettings: state.currentGame.globalSettings,
      // State
      isRunning: state.currentGame.isRunning,
      isStarted: state.currentGame.isStarted,
      isFinished: state.currentGame.isFinished,
      animationInProgress: state.currentGame.animationInProgress,
      // Base data
      board: state.currentGame.board,
      // Game data
      maxMovesCount: state.currentGame.maxMovesCount,
      movesCount: state.currentGame.movesCount,
      progress: state.currentGame.progress,
      progressTotal: state.currentGame.progressTotal,
      progressDelta: state.currentGame.progressDelta,
      gameWon: state.currentGame.gameWon,
    );
    // game.dump();

    updateState(game);
  }

  void startNewGame({
    required GameSettings gameSettings,
    required GlobalSettings globalSettings,
  }) {
    final Game newGame = Game.createNew(
      // Settings
      gameSettings: gameSettings,
      globalSettings: globalSettings,
    );

    newGame.dump();

    updateState(newGame);
    updateGameIsRunning(true);

    refresh();
  }

  void quitGame() {
    state.currentGame.isRunning = false;
    refresh();
  }

  void resumeSavedGame() {
    state.currentGame.isRunning = true;
    refresh();
  }

  void deleteSavedGame() {
    state.currentGame.isRunning = false;
    state.currentGame.isFinished = true;
    refresh();
  }

  void updateCellValue(int col, int row, int value) {
    state.currentGame.board.cells[row][col].value = value;
    refresh();
  }

  void increaseMovesCount() {
    state.currentGame.movesCount++;
    refresh();
  }

  void updateProgressDelta(int progressDelta) {
    state.currentGame.progressDelta = progressDelta;
    refresh();
  }

  void updateProgress(int progress) {
    state.currentGame.progress = progress;
    refresh();
  }

  void updateAnimationInProgress(bool inProgress) {
    state.currentGame.animationInProgress = inProgress;
    refresh();
  }

  void updateGameIsRunning(bool gameIsRunning) {
    state.currentGame.isRunning = gameIsRunning;
    refresh();
  }

  void updateGameIsStarted(bool gameIsStarted) {
    state.currentGame.isStarted = gameIsStarted;
    refresh();
  }

  void updateGameIsFinished(bool gameIsFinished) {
    state.currentGame.isFinished = gameIsFinished;
    refresh();
  }

  void updateGameWon(bool gameWon) {
    state.currentGame.gameWon = gameWon;
    refresh();
  }

  fillBoardFromFirstCell(int value) {
    updateGameIsStarted(true);

    List<List<int>> cellsToFill = state.currentGame.board.getSiblingFillableCells(0, 0, []);
    final int progressBeforeMove = cellsToFill.length;

    if (value == state.currentGame.board.getFirstCellValue()) {
      return;
    }

    increaseMovesCount();

    // Sort cells from the closest to the furthest, relatively to the top left corner
    cellsToFill
        .sort((a, b) => (pow(a[0], 2) + pow(a[1], 2)).compareTo(pow(b[0], 2) + pow(b[1], 2)));

    const interval = Duration(milliseconds: 10);
    int cellIndex = 0;
    updateAnimationInProgress(true);
    Timer.periodic(
      interval,
      (Timer timer) {
        if (cellIndex < cellsToFill.length) {
          updateCellValue(cellsToFill[cellIndex][1], cellsToFill[cellIndex][0], value);
          cellIndex++;
        } else {
          timer.cancel();

          int progressAfterMove =
              state.currentGame.board.getSiblingFillableCells(0, 0, []).length;
          int progressDelta = progressAfterMove - progressBeforeMove;
          updateProgressDelta(progressDelta);
          updateProgress(progressAfterMove);

          updateAnimationInProgress(false);

          if (state.currentGame.board.isBoardSolved()) {
            updateGameWon(true);
            updateGameIsFinished(true);
          }
        }
      },
    );
  }

  @override
  GameState? fromJson(Map<String, dynamic> json) {
    final Game currentGame = json['currentGame'] as Game;

    return GameState(
      currentGame: currentGame,
    );
  }

  @override
  Map<String, dynamic>? toJson(GameState state) {
    return <String, dynamic>{
      'currentGame': state.currentGame,
    };
  }
}
