import 'dart:math';

import 'package:flutter/material.dart';

import 'package:colors/config/color_theme.dart';
import 'package:colors/config/default_game_settings.dart';
import 'package:colors/config/default_global_settings.dart';
import 'package:colors/models/settings/settings_game.dart';
import 'package:colors/models/settings/settings_global.dart';
import 'package:colors/utils/color_extensions.dart';
import 'package:colors/utils/tools.dart';

class ParameterPainter extends CustomPainter {
  const ParameterPainter({
    required this.code,
    required this.value,
    required this.isSelected,
    required this.gameSettings,
    required this.globalSettings,
  });

  final String code;
  final String value;
  final bool isSelected;
  final GameSettings gameSettings;
  final GlobalSettings globalSettings;

  @override
  void paint(Canvas canvas, Size size) {
    // force square
    final double canvasSize = min(size.width, size.height);

    const Color borderColorEnabled = Colors.blue;
    const Color borderColorDisabled = Colors.white;

    // "enabled/disabled" border
    final paint = Paint();
    paint.style = PaintingStyle.stroke;
    paint.color = isSelected ? borderColorEnabled : borderColorDisabled;
    paint.strokeJoin = StrokeJoin.round;
    paint.strokeWidth = 10;
    canvas.drawRect(
        Rect.fromPoints(const Offset(0, 0), Offset(canvasSize, canvasSize)), paint);

    // content
    switch (code) {
      case DefaultGameSettings.parameterCodeDifficultyLevel:
        paintDifficultyLevelParameterItem(value, canvas, canvasSize);
        break;
      case DefaultGameSettings.parameterCodeBoardSize:
        paintBoardSizeParameterItem(value, canvas, canvasSize);
        break;
      case DefaultGameSettings.parameterCodeColorsCount:
        paintColorsCountParameterItem(value, canvas, canvasSize);
        break;
      case DefaultGlobalSettings.parameterCodeSkin:
        paintUnknownParameterItem(value, canvas, canvasSize);
        break;
      default:
        printlog('Unknown parameter: $code/$value');
        paintUnknownParameterItem(value, canvas, canvasSize);
    }
  }

  @override
  bool shouldRepaint(CustomPainter oldDelegate) {
    return false;
  }

  // "unknown" parameter -> simple block with text
  void paintUnknownParameterItem(
    final String value,
    final Canvas canvas,
    final double size,
  ) {
    final paint = Paint();
    paint.strokeJoin = StrokeJoin.round;
    paint.strokeWidth = 3;

    paint.color = Colors.grey;
    paint.style = PaintingStyle.fill;
    canvas.drawRect(Rect.fromPoints(const Offset(0, 0), Offset(size, size)), paint);

    final textSpan = TextSpan(
      text: '?\n$value',
      style: const TextStyle(
        color: Colors.black,
        fontSize: 18,
        fontWeight: FontWeight.bold,
      ),
    );
    final textPainter = TextPainter(
      text: textSpan,
      textDirection: TextDirection.ltr,
      textAlign: TextAlign.center,
    );
    textPainter.layout();
    textPainter.paint(
      canvas,
      Offset(
        (size - textPainter.width) * 0.5,
        (size - textPainter.height) * 0.5,
      ),
    );
  }

  void paintDifficultyLevelParameterItem(
    final String value,
    final Canvas canvas,
    final double size,
  ) {
    Color backgroundColor = Colors.grey;

    final List<dynamic> stars = [];

    switch (value) {
      case DefaultGameSettings.difficultyLevelValueEasy:
        backgroundColor = Colors.green;
        stars.add([0.5, 0.5]);
        break;
      case DefaultGameSettings.difficultyLevelValueMedium:
        backgroundColor = Colors.orange;
        stars.add([0.3, 0.5]);
        stars.add([0.7, 0.5]);
        break;
      case DefaultGameSettings.difficultyLevelValueHard:
        backgroundColor = Colors.red;
        stars.add([0.3, 0.3]);
        stars.add([0.7, 0.3]);
        stars.add([0.5, 0.7]);
        break;
      case DefaultGameSettings.difficultyLevelValueNightmare:
        backgroundColor = Colors.purple;
        stars.add([0.3, 0.3]);
        stars.add([0.7, 0.3]);
        stars.add([0.3, 0.7]);
        stars.add([0.7, 0.7]);
        break;
      default:
        printlog('Wrong value for level parameter value: $value');
    }

    final paint = Paint();
    paint.strokeJoin = StrokeJoin.round;
    paint.strokeWidth = 3 / 100 * size;

    // Colored background
    paint.color = backgroundColor;
    paint.style = PaintingStyle.fill;
    canvas.drawRect(Rect.fromPoints(const Offset(0, 0), Offset(size, size)), paint);

    // Stars
    final textSpan = TextSpan(
      text: '⭐',
      style: TextStyle(
        color: Colors.black,
        fontSize: size / 3,
        fontWeight: FontWeight.bold,
      ),
    );
    final textPainter = TextPainter(
      text: textSpan,
      textDirection: TextDirection.ltr,
      textAlign: TextAlign.center,
    );
    textPainter.layout();

    for (var center in stars) {
      textPainter.paint(
        canvas,
        Offset(
          size * center[0] - textPainter.width * 0.5,
          size * center[1] - textPainter.height * 0.5,
        ),
      );
    }
  }

  void paintBoardSizeParameterItem(
    final String value,
    final Canvas canvas,
    final double size,
  ) {
    Color backgroundColor = Colors.grey;
    int gridWidth = 1;

    switch (value) {
      case DefaultGameSettings.boardSizeValueSmall:
        backgroundColor = Colors.green;
        gridWidth = 2;
        break;
      case DefaultGameSettings.boardSizeValueMedium:
        backgroundColor = Colors.orange;
        gridWidth = 3;
        break;
      case DefaultGameSettings.boardSizeValueLarge:
        backgroundColor = Colors.red;
        gridWidth = 4;
        break;
      case DefaultGameSettings.boardSizeValueExtra:
        backgroundColor = Colors.purple;
        gridWidth = 5;
        break;
      default:
        printlog('Wrong value for boardSize parameter value: $value');
    }

    final paint = Paint();
    paint.strokeJoin = StrokeJoin.round;
    paint.strokeWidth = 3 / 100 * size;

    // Colored background
    paint.color = backgroundColor;
    paint.style = PaintingStyle.fill;
    canvas.drawRect(Rect.fromPoints(const Offset(0, 0), Offset(size, size)), paint);

    // Mini grid
    final squareBackgroundColor = Colors.grey.shade200;
    final squareBorderColor = Colors.grey.shade800;

    final double cellSize = size / 7;
    final double origin = (size - gridWidth * cellSize) / 2;

    for (int row = 0; row < gridWidth; row++) {
      for (int col = 0; col < gridWidth; col++) {
        final Offset topLeft = Offset(origin + col * cellSize, origin + row * cellSize);
        final Offset bottomRight = topLeft + Offset(cellSize, cellSize);

        paint.color = squareBackgroundColor;
        paint.style = PaintingStyle.fill;
        canvas.drawRect(Rect.fromPoints(topLeft, bottomRight), paint);

        paint.color = squareBorderColor;
        paint.style = PaintingStyle.stroke;
        canvas.drawRect(Rect.fromPoints(topLeft, bottomRight), paint);
      }
    }
  }

  void paintColorsCountParameterItem(
    final String value,
    final Canvas canvas,
    final double size,
  ) {
    Color backgroundColor = Colors.grey;

    switch (value) {
      case DefaultGameSettings.colorsCountValueLow:
        backgroundColor = Colors.green;
        break;
      case DefaultGameSettings.colorsCountValueMedium:
        backgroundColor = Colors.orange;
        break;
      case DefaultGameSettings.colorsCountValueHigh:
        backgroundColor = Colors.red;
        break;
      case DefaultGameSettings.colorsCountValueVeryHigh:
        backgroundColor = Colors.purple;
        break;
      default:
        printlog('Wrong value for colorsCount parameter value: $value');
    }

    final paint = Paint();
    paint.strokeJoin = StrokeJoin.round;
    paint.strokeWidth = 3;

    // Colored background
    paint.color = backgroundColor;
    paint.style = PaintingStyle.fill;
    canvas.drawRect(Rect.fromPoints(const Offset(0, 0), Offset(size, size)), paint);

    // Colors preview
    const List<Offset> positions = [
      Offset(0, 0),
      Offset(1, 0),
      Offset(2, 0),
      Offset(2, 1),
      Offset(2, 2),
      Offset(1, 2),
      Offset(0, 2),
      Offset(0, 1),
    ];

    const skin = DefaultGlobalSettings.defaultSkinValue;

    final double padding = 4 / 100 * size;
    final double margin = 3 / 100 * size;
    final double width = ((size - 2 * padding) / 3) - 2 * margin;

    final int maxValue = int.parse(value);
    for (int colorIndex = 0; colorIndex < maxValue; colorIndex++) {
      final Offset position = positions[colorIndex];

      final Offset topLeft = Offset(padding + margin + position.dx * (width + 2 * margin),
          padding + margin + position.dy * (width + 2 * margin));

      final Offset bottomRight = topLeft + Offset(width, width);

      final squareColor = Color(ColorTheme.getColorCode(skin, colorIndex + 1));
      paint.color = squareColor;
      paint.style = PaintingStyle.fill;
      canvas.drawRect(Rect.fromPoints(topLeft, bottomRight), paint);

      final borderColor = squareColor.darken(20);
      paint.color = borderColor;
      paint.style = PaintingStyle.stroke;
      canvas.drawRect(Rect.fromPoints(topLeft, bottomRight), paint);
    }

    // centered text value
    final textSpan = TextSpan(
      text: value.toString(),
      style: TextStyle(
        color: Colors.black,
        fontSize: size / 4,
        fontWeight: FontWeight.bold,
      ),
    );
    final textPainter = TextPainter(
      text: textSpan,
      textDirection: TextDirection.ltr,
      textAlign: TextAlign.center,
    );
    textPainter.layout();
    textPainter.paint(
      canvas,
      Offset(
        (size - textPainter.width) * 0.5,
        (size - textPainter.height) * 0.5,
      ),
    );
  }

  void paintColorsThemeParameterItem(
    final String value,
    final Canvas canvas,
    final double size,
  ) {
    Color backgroundColor = Colors.grey;
    const int gridWidth = 4;

    final paint = Paint();
    paint.strokeJoin = StrokeJoin.round;
    paint.strokeWidth = 3;

    // Colored background
    paint.color = backgroundColor;
    paint.style = PaintingStyle.fill;
    canvas.drawRect(Rect.fromPoints(const Offset(0, 0), Offset(size, size)), paint);

    // Mini grid
    final borderColor = Colors.grey.shade800;

    final double cellSize = size / gridWidth;
    final double origin = (size - gridWidth * cellSize) / 2;

    for (int row = 0; row < gridWidth; row++) {
      for (int col = 0; col < gridWidth; col++) {
        final Offset topLeft = Offset(origin + col * cellSize, origin + row * cellSize);
        final Offset bottomRight = topLeft + Offset(cellSize, cellSize);

        const squareColor = Colors.pink;

        paint.color = squareColor;
        paint.style = PaintingStyle.fill;
        canvas.drawRect(Rect.fromPoints(topLeft, bottomRight), paint);

        paint.color = borderColor;
        paint.style = PaintingStyle.stroke;
        canvas.drawRect(Rect.fromPoints(topLeft, bottomRight), paint);
      }
    }
  }
}
